# the actual API
import warnings

warnings.filterwarnings("ignore")

import matplotlib.pyplot as plt

plt.rcParams["pdf.fonttype"] = 42
plt.rcParams["svg.fonttype"] = "none"

from trackc import pa, pl, tl
from trackc.gs import make_spec, savefig, tenon

__all__ = ["tl", "pl", "pa", "make_spec", "savefig", "tenon"]



import argparse


class _genebed12(object):
    """bed12 object"""

    def __init__(self, chrom, start, end, name, strand, gene_biotype):
        self.chrom = chrom
        self.start = start
        self.end = end
        self.name = name
        self.score = 0
        self.strand = strand
        self.thickStart = end
        self.thickEnd = end
        self.itemRgb = 0
        self.blockCount = 0
        self.exon_site = {}
        self.gene_biotype = gene_biotype
        # self.blockSizes = {}
        # self.blockStarts = {}

    def add_exon_block(self, s, e):
        # self.blockCount += 1
        self.exon_site["{0}-{1}".format(s, e)] = 0

    def tostr(self, biotype2bed13):
        blockSizes = None
        blockStarts = None
        blockCount = len(self.exon_site)

        for i, v in self.exon_site.items():
            tmp = i.split("-")
            if blockSizes == None:
                blockSizes = str(int(tmp[1]) - int(tmp[0]))
                blockStarts = tmp[0]
            else:
                blockSizes = blockSizes + "," + "{0}".format(int(tmp[1]) - int(tmp[0]))
                blockStarts = blockStarts + "," + tmp[0]

        obj_str = "{0}\t{1}\t{2}\t{3}\t{4}\t{5}\t{6}\t{7}\t{8}\t{9}\t{10}\t{11}".format(
            self.chrom,
            self.start,
            self.end,
            self.name,
            self.score,
            self.strand,
            self.thickStart,
            self.thickEnd,
            self.itemRgb,
            blockCount,
            blockSizes,
            blockStarts,
            self.gene_biotype,
        )
        if biotype2bed13 == True:
            obj_str = "{0}\t{1}\t{2}\t{3}\t{4}\t{5}\t{6}\t{7}\t{8}\t{9}\t{10}\t{11}\t{12}".format(
                self.chrom,
                self.start,
                self.end,
                self.name,
                self.score,
                self.strand,
                self.thickStart,
                self.thickEnd,
                self.itemRgb,
                blockCount,
                blockSizes,
                blockStarts,
            )
        return obj_str


def _bedcol9_2dic(bedcol9, split=";"):
    tmp = bedcol9.split(split)
    col9dic = {}
    for i in tmp:
        i = i.strip()
        ii = i.split(" ")
        if len(ii) < 2:
            continue
        val = ii[1].strip('"')
        col9dic[ii[0]] = val
    return col9dic


def gtf2bed(
    gtf, outfile, sep, gene_name_tag, gene_id_tag, gene_biotype_tag, biotype2bed13
):
    gene = None
    gtf_hand = open(gtf, "r")
    gtf2bed = open(outfile, "w")

    for line in gtf_hand:
        bedtab = line.rstrip().split("\t")
        if len(bedtab) < 9:
            continue

        # if re.match('chr', bedtab[0]) == None:
        #    continue

        # if re.match('chrM', bedtab[0]) != None:
        #    continue

        # print(bedtab)
        # print(gene)

        if bedtab[2] == "gene":
            if gene != None:
                gtf2bed.write(gene.tostr(biotype2bed13) + "\n")
            Gene_name = "xx"
            Gene_biotype = "yy"

            col9dic = _bedcol9_2dic(bedtab[8], split=sep)
            if gene_name_tag in col9dic.keys():
                Gene_name = col9dic[gene_name_tag]
            elif gene_id_tag in col9dic.keys():
                Gene_name = col9dic[gene_id_tag]
            else:
                print(bedtab[8], " have no {0}, xx instead".format(gene_name_tag))

            if gene_biotype_tag in col9dic.keys():
                Gene_biotype = col9dic[gene_biotype_tag]
            else:
                print(bedtab[8], " have no {0}, yy instead".format(gene_biotype_tag))

            gene = _genebed12(
                bedtab[0], bedtab[3], bedtab[4], Gene_name, bedtab[6], Gene_biotype
            )

        elif bedtab[2] == "exon":
            if gene != None:
                gene.add_exon_block(bedtab[3], bedtab[4])
        else:
            pass

    # 最后一行没有输出

    gtf_hand.close()
    gtf2bed.close()
    print("gtf2bed finished")



#!/Users/yuanzan/anaconda3/bin/python3
# -*- coding: utf-8 -*-

import warnings

import click
import yaml

warnings.filterwarnings("ignore")

import trackc as tc
from trackc.scripts.gtf2bed12 import gtf2bed
from trackc.scripts.trackc_cli import cli as track

# from trackc_cli import track
# from trackc.scripts.tadscore import tadScore


@click.group()
def main():
    pass


@main.command(
    name="cli",
    short_help="plot trackc by use yaml format configfile, yaml track type parameters is the same as API",
)
@click.argument("config", metavar="<trackc-conf.yml>")
@click.option(
    "--regions",
    "-r",
    default=None,
    help="genome regions, \
              eg.:'18:47950000-48280000 18:75280000-74850000'",
)
@click.option(
    "--outfile", "-o", default="trackc.pdf", help="output filename. default=trackc.pdf"
)
@click.option(
    "--basefigsize",
    "-s",
    default="6,1",
    help='base figsize: width,height. default="6,1"\
              The height option in the config.yml is relative to the base figsize height',
)
def cli(config, regions, outfile, basefigsize):
    track(config, regions, outfile, basefigsize)


@main.command(
    name="gtf2bed", short_help="Converts the gene GTF to bed12 or bed13 format"
)
@click.argument("gtf", metavar="<genome.gtf>")
# , help="input file. GTF format gene annotation."
@click.option(
    "--outfile",
    "-o",
    default="gene.bed",
    help="File name of gene annotation bed12 format.",
)
@click.option("--sep", "-s", default=";", help="tags of GTF sepration by which str")
@click.option(
    "--gene_name_tag", "-nt", default="gene_name", help="tags of GTF gene_name"
)
@click.option("--gene_id_tag", "-it", default="gene_id", help="tags of GTF gene_id")
@click.option(
    "--gene_biotype_tag", "-bt", default="gene_biotype", help="tags of GTF gene_biotype"
)
@click.option(
    "--biotype2bed13",
    "-bed13",
    is_flag=True,
    show_default=True,
    default=True,
    help="if this arg is set, the out file columns 13 will add gene's biotype",
)
def mode2(
    gtf, outfile, sep, gene_name_tag, gene_id_tag, gene_biotype_tag, biotype2bed13
):
    gtf2bed(
        gtf, outfile, sep, gene_name_tag, gene_id_tag, gene_biotype_tag, biotype2bed13
    )


# @main.command()
# @click.argument('mcool', metavar='name.mcool::/resolutions/10000')

# def tadscore(mcool):
#    tadScore()


if __name__ == "__main__":
    main()



import sys

import pandas as pd
import yaml

import trackc as tc


def _get_yaml_data(yaml_file):
    file = open(yaml_file, "r", encoding="utf-8")
    file_data = file.read()
    file.close()

    data = yaml.safe_load(file_data)
    return data


def init_ax(ax_dic, height=1, hspace=0.06):
    if "height" in ax_dic.keys():
        if ax_dic["height"] != None:
            height = float(ax_dic["height"])

    if "hspace" in ax_dic.keys():
        if ax_dic["hspace"] != None:
            hspace = float(ax_dic["hspace"])

    return height, hspace


def _make_spec(conf):
    df = pd.DataFrame({"ax": [tt["ax"] for i, tt in enumerate(conf["trackc"])]})
    df["track_type"] = [tt["track_type"] for i, tt in enumerate(conf["trackc"])]
    axs = pd.DataFrame(
        {"rank_u": range(len(df["ax"].unique()))}, index=df["ax"].unique()
    )
    axs["height"] = 1
    axs["hspace"] = 0.05
    axs["bottom"] = "bottom"

    for i, tt in enumerate(conf["trackc"]):
        height, hspace = init_ax(
            tt, axs.loc[tt["ax"], "height"], axs.loc[tt["ax"], "hspace"]
        )
        axs.loc[tt["ax"], "height"] = height
        axs.loc[tt["ax"], "hspace"] = hspace

    return df, axs


def hicContactMap(Paras, ax):
    mapc_paras = Paras["mapC"]
    mat = None
    mat2 = None
    if "mat" in Paras.keys():
        if Paras["mat"]["method"] == "extractCisContact":
            del Paras["mat"]["method"]
            mat = tc.tl.extractCisContact(**Paras["mat"])

        elif Paras["mat"]["method"] == "extractContactRegions":
            if (
                "row_regions" in Paras["mat"].keys()
                or "col_regions" in Paras["mat"].keys()
            ):
                del Paras["mat"]["method"]
                mat_obj = tc.tl.extractContactRegions(
                    clr=Paras["mat"]["clr"], row_regions=region
                )
                mat = mat_obj.cmat
        else:
            pass

    if "mat2" in Paras.keys():
        if Paras["mat2"]["method"] == "extractCisContact":
            del Paras["mat2"]["method"]
            mat2 = tc.tl.extractCisContact(**Paras["mat2"])

        elif Paras["mat2"]["method"] == "extractContactRegions":
            if (
                "row_regions" in Paras["mat2"].keys()
                or "col_regions" in Paras["mat2"].keys()
            ):
                del Paras["mat2"]["method"]
                mat_obj = tc.tl.extractContactRegions(
                    clr=Paras["mat2"]["clr"], row_regions=region
                )
                mat2 = mat_obj.cmat
        else:
            pass

    if mat is None:
        tc.pl.mapC(ax=ax, mat2=mat2, **Paras["mapC"])
    else:
        if mat2 is not None:
            tc.pl.mapC(ax=ax, mat=mat, mat2=mat2, **Paras["mapC"])
        else:
            tc.pl.mapC(ax=ax, mat=mat, **Paras["mapC"])


"""

@click.command()
@click.argument("config", metavar="<trackc-conf.yml>")
@click.option(
    "--regions",
    "-r",
    default=None,
    help="genome regions, \
              eg.:'18:47950000-48280000 18:75280000-74850000'",
)
@click.option(
    "--outfile", "-o", default="trackc.pdf", help="output filename. default=trackc.pdf"
)
@click.option(
    "--basefigsize",
    "-s",
    default="6,1",
    help='base figsize: width,height. default="6,1"\
              The height option in the config.yml is relative to the base figsize height',
)
"""


def cli(config, regions, outfile, basefigsize):
    conf = _get_yaml_data(config)
    fs = basefigsize.split(",")
    bfs = (float(fs[0]), float(fs[1]))
    if regions != None:
        region_tmp = regions.split(" ")
        if len(region_tmp) > 1:
            regions = region_tmp

    # make grid spec
    df, axs_pre = _make_spec(conf)

    ten = tc.tenon(figsize=bfs)
    for i, row in axs_pre.iterrows():
        ten.add(pos=row["bottom"], height=row["height"], hspace=row["hspace"])

    regions_type = [
        "multi_scale_track",
        "bed_track",
        "bw_track",
        "virtual4C",
        "links_track",
        "gene_track",
        "bdgmat_track",
    ]
    region_type = ["scale_track"]

    for i, tt in enumerate(conf["trackc"]):
        track_type = tt["track_type"]
        if track_type == None:
            ax = ten.axs(axs_pre.loc[tt["ax"], "rank_u"])
            ax.set_axis_off()
            continue

        print("ax: ", tt["ax"])

        paras = tt["track_para"]
        if track_type in regions_type:
            if "regions" not in paras.keys():
                if regions != None:
                    paras["regions"] = regions
                else:
                    print("region not set")
                    sys.exit(1)

        if track_type in region_type:
            if "region" not in paras.keys():
                if regions != None:
                    paras["region"] = regions
                else:
                    # print('region para not set in cli and conf.yml')
                    pass

        if track_type == "hicmap":
            if "mat" in paras.keys():
                if paras["mat"]["method"] == "extractCisContact":
                    if "region" not in paras["mat"].keys():
                        paras["mat"]["region"] = regions
                elif paras["mat"]["method"] == "extractContactRegions":
                    if (
                        "row_regions" not in paras["mat"].keys()
                        and "col_regions" not in paras["mat"].keys()
                    ):
                        paras["mat"]["row_regions"] = regions
                else:
                    pass

            if "mat2" in paras.keys():
                if paras["mat2"]["method"] == "extractCisContact":
                    if "region" not in paras["mat2"].keys():
                        paras["mat2"]["region"] = regions
                elif paras["mat2"]["method"] == "extractContactRegions":
                    if (
                        "row_regions" not in paras["mat2"].keys()
                        and "col_regions" not in paras["mat2"].keys()
                    ):
                        paras["mat2"]["row_regions"] = regions
                else:
                    pass

        ax = ten.axs(axs_pre.loc[tt["ax"], "rank_u"])

        if track_type == "multi_scale_track":
            tc.pl.multi_scale_track(**paras, ax=ax)

        if track_type == "scale_track":
            tc.pl.scale_track(**paras, ax=ax)

        if track_type == "bed_track":
            tc.pl.bed_track(**paras, ax=ax)

        if track_type == "bw_track":
            tc.pl.bw_track(**paras, ax=ax)

        if track_type == "virtual4C":
            tc.pl.virtual4C(**paras, ax=ax)

        if track_type == "links_track":
            tc.pl.links_track(**paras, ax=ax)

        if track_type == "gene_track":
            tc.pl.gene_track(**paras, ax=ax)

        if track_type == "hicmap":
            hicContactMap(paras, ax=ax)

        if track_type == "zoomin":
            tc.pl.zoomin(**paras, ax=ax)

        if track_type == "bdgmat_track":
            tc.pl.bdgmat_track(**paras, ax=ax)

    tc.savefig(outfile)



from typing import Optional, Sequence, Union

import cooler
import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib.axes import Axes
from matplotlib.colors import Colormap

from trackc.pl.mapc import getData2Map, mapC
from trackc.tl._getRegionsCmat import extractContactRegions

fruitpunch = sns.blend_palette(["white", "red"], as_cmap=True)


def _plot4C_line_bar(
    data,
    ax,
    color,
    track_type,
    logdata,
    trim_range,
    minrange,
    maxrange,
):
    data, maxrange, minrange = getData2Map(
        data, maxrange=maxrange, minrange=minrange, trim_range=trim_range, inplace=False
    )
    print("maxrange:", maxrange, "minrange:", minrange)

    if logdata:
        data = np.log2(data)

    if track_type == "line":
        ax.plot(range(len(data[0, :])), data[0, :], color=color, solid_capstyle="butt")

    if track_type == "bar":
        ax.bar(
            x=range(len(data[0, :])),
            height=data[0, :],
            bottom=minrange,
            width=1,
            color=color,
            align="edge",
        )
    return (data, maxrange, minrange)


def _get_pets(df, binsize=10000):
    pets = df.reset_index()
    pets_df = pd.DataFrame(index=range(pets["cbins"].sum()))
    pets_df["chrom"] = "None"
    pets_df["bin_N"] = 0

    chroms = []
    bins = []

    for i, row in pets.iterrows():
        chroms.extend([row["chrom"]] * row["cbins"])

        bin_s = int(row["fetch_start"] / binsize)
        # if row['fetch_start']%binsize > 0:
        #    bin_s = bin_s

        bin_e = int(row["fetch_end"] / binsize)
        if row["fetch_end"] % binsize > 0:
            bin_e = bin_e + 1

        chrbins = list(range(bin_s, bin_e))

        if row["isReverse"] == True:
            chrbins = chrbins[::-1]
        bins.extend(chrbins)

    pets_df["chrom"] = chroms
    pets_df["chrbin"] = bins
    return pets_df


def virtual4C(
    clr: Union[cooler.Cooler, str],
    ax: Optional[Axes] = None,
    balance: bool = False,
    # divisive_weights = None,
    target: Union[str, None] = None,
    regions: Union[Sequence[str], str, None] = None,
    track_type: Union[str, None] = "line",
    color: Union[str, None] = "tab:blue",
    cmap: Union[Sequence[Colormap], str, None] = fruitpunch,
    target_color: Union[str, None] = None,
    target_name: Union[str, None] = None,
    logdata: bool = False,
    trim_range: float = 0.98,
    minrange: float = None,
    maxrange: float = None,
    label: Optional[str] = None,
    label_rotation: Union[int, None] = 0,
    label_fontsize: Optional[int] = 12,
):
    """
    Plot virtual4C track, support for multiple or reverse genome regions.

    Parameters
    ----------
    ax: :class:`matplotlib.axes.Axes` object
    clr: `cooler.Cooler` | `str`
        cool format Hi-C matrix (https://github.com/open2c/cooler) or cool file path
        `cooler.Cooler` e.g. GM12878=cooler.Cooler('./GM12878.chr18.mcool::/resolutions/50000')
        `str` e.g. 'GM12878.chr18.mcool::/resolutions/50000' or 'GM12878.chr18.cool'
    balance: `bool`
        The ``'balance'`` parameters of ``coolMat.matrix(balance=False).fetch('chr6:119940450-123940450')``
    target: `str`
        Coordinates of the viewpoint. e.g. 'chr8:127735434-127735435'
    regions: `str` | `str list`
        The genome regions, which contact to the viewpoint
        e.g. ``"chr6:1000000-2000000"`` or ``["chr6:1000000-2000000", "chr3:5000000-4000000", "chr5"]``
        The start can be larger than the end (eg. ``"chr6:2000000-1000000"``),
        which means you want to get the reverse region contacts
    track_type: `str`
        virtual4C types, you can choose one of ['line', 'bar', 'heatmap']
    color: `str`
        `line` or `bar` type virtual4C color
    cmap: `str` | `matplotlib.colors.Colormap`
        if track_type set `heatmap`, the cmap
    target_color: `str`
        virtual4C viewpoint color
    target_name: `str`
        viewpoint label
    logdata: `bool`
        do you want to log the data before plotting
    trim_range: `float`
        remove the extreme values by trimming the counts.[0,1]
    minrange: `float`
        the minimum range of values used to define the ylim or color palette
    maxrange: `float`
        the maximum range of values used to define the ylim or color palette
    label: `str`
        the title of the track, will show on the left
    label_rotation: `int`
        the label text rotation
    label_fontsize: `int`
        the label text fontsize

    Example
    -------
    >>> import trackc as tc
    >>> regions = ['chr8:127000000-129200000', 'chr14:96500000-99300000']
    >>> MYC_TSS = 'chr8:127735434-127735435'
    >>> ten = tc.tenon(figsize=(8,1))
    >>> ten.add(pos='bottom', height=1, hspace=0.1)
    >>> AML_1360 = cooler.Cooler('./GSM4604287_1360.iced.mcool::/resolutions/10000')
    >>> tc.pl.virtual4C(ax=ten.axs(0), clr=AML_1360, target=MYC_TSS, regions=regions,
                track_type='line', label='Virtual 4C', target_color='r')
    >>> tc.savefig('trackc_virtual4c.pdf')
    """
    if isinstance(clr, str):
        clr = cooler.Cooler(clr)

    data = extractContactRegions(
        clr, balance=balance, row_regions=target, col_regions=regions
    )
    cols = _get_pets(data.col_regions, clr.binsize)
    rows = _get_pets(data.row_regions, clr.binsize)
    cols["target"] = 0
    cols.loc[cols["chrbin"].isin(rows["chrbin"]), "target"] = 1
    targets_point = cols.query("target==1")

    target_x = 0
    target_height = 0
    target_bottom = 0

    if track_type in ["line", "bar"]:
        _, maxrange, minrange = _plot4C_line_bar(
            data=data.cmat,
            ax=ax,
            color=color,
            track_type=track_type,
            logdata=logdata,
            trim_range=trim_range,
            minrange=minrange,
            maxrange=maxrange,
        )

        # ax.bar_label(target_bar, label_type='edge')
        ax.text(
            targets_point.index[0],
            minrange + (maxrange - minrange) / 2,
            target_name,
            va="center",
            ha="right",
            rotation=90,
        )
        ax.set_ylim(minrange, maxrange)
        ax.set_xlim(0, data.cmat.shape[1] - 1)

        target_x = targets_point.index[0]
        target_height = maxrange
        target_bottom = minrange

    if track_type == "heatmap":
        mapC(
            ax=ax,
            mat=data.cmat,
            cmap=cmap,
            logdata=logdata,
            symmetric=True,
            trim_range=trim_range,
            minrange=minrange,
            maxrange=maxrange,
            map_type="square",
            ax_on=True,
            aspect="auto",
        )
        ax.set_aspect(aspect="auto")

        target_bottom = -0.5
        target_height = 1
        target_x = targets_point.index[0] - 0.5

    target_bar = ax.bar(
        x=target_x,
        height=target_height,
        bottom=target_bottom,
        width=1,
        color=target_color,
        align="edge",
        label=target_name,
    )

    ax.tick_params(bottom=False, top=False, left=False, right=True)
    ax.yaxis.tick_right()
    # ax.set_xticklabels("")
    ax.set_xticklabels("")
    ax.set_ylabel(
        label, fontsize=label_fontsize, rotation=label_rotation, ha="right", va="center"
    )



from typing import Optional, Sequence, Union

import pandas as pd
from matplotlib.axes import Axes

from trackc.pa import trackcl_11
from trackc.tl import GenomeRegion


def _scale_ticks(ticks, scale="bp", tick_fl="%0.2f"):
    if scale == "Mb":
        ticks = ticks / 1000000
        ticks = ["{0}".format(tick_fl % i) for i in ticks]
    elif scale == "kb":
        ticks = ticks / 1000
        ticks = ["{0}".format(tick_fl % i) for i in ticks]
    else:
        return ticks
    return ticks


def scale_track(
    ax: Optional[Axes] = None,
    region: Union[str, None] = None,
    tick_pos: str = "bottom",
    ratio2ax: float = 0.5,
    label_fontsize: Union[int, None] = 10,
    scale_adjust: Union[str, None] = "kb",
    tick_fl: Union[str, None] = "%0.2f",
    tick_fontsize: Union[int, None] = 8,
    tick_rotation: Union[int, None] = 0,
    space: float = 0.1,
):
    """
    Plot one region scale bar track

    Parameters
    ----------
    ax: :class:`matplotlib.axes.Axes` object
    region: `str`
        one genome region, format: `chrom:start-end`.
        e.g. ``"chr18:45000000-78077248"``

        if the start is bigger than end, the genome region will be reversed
    tick_pos: `str`
        ticks position, can be one of ['top', 'bottom']
    ratio2ax: `float`
        the height ratio refer to the given `ax`'s height
    label_fontsize: `int`
        the region text fontsize
    scale_adjust: `str`
        adjust the scale unit to make it pretty, can be one of ['kb', 'Mb']
    tick_fl: `str`
        ticks retains a few decimal places
    tick_fontsize: `int`
        ticks text fontsize
    tick_rotation: `int`
        ticks text rotation
    space: `float`
        space relative to the ax

    Example
    -------
    >>> import trackc as tc
    >>> region = 'chr7:153000000-151000000'

    >>> ten = tc.tenon(figsize=(8,1))
    >>> ten.add(pos='bottom', height=0.5)
    >>> ten.axs(0).axis('off')

    >>> tc.pl.scale_track(ax=ten.axs(0), region=region, scale_adjust='Mb', tick_pos='bottom', ratio2ax=1.2)
    >>> tc.pl.scale_track(ax=ten.axs(0), region=region, scale_adjust='Mb', tick_pos='top', ratio2ax=1.2)
    >>> tc.savefig('trackc_scalebar_track.pdf')
    """

    line_GenomeRegions = None
    if isinstance(region, list):
        print("scale_track is only for one region")
        return
    else:
        line_GenomeRegions = GenomeRegion(region).GenomeRegion2df()

    line_GenomeRegions["raw_region"] = line_GenomeRegions.index
    line_GenomeRegions = line_GenomeRegions.reset_index()

    chrom = line_GenomeRegions.loc[0, "chrom"]
    start = line_GenomeRegions.loc[0, "start"]
    end = line_GenomeRegions.loc[0, "end"]
    raw_region = line_GenomeRegions.loc[0, "raw_region"]

    pos_dic = {
        "left": [-ratio2ax, 0, ratio2ax, 1],
        "right": [1, 0, ratio2ax, 1],
        "top": [0, 1 + space, 1, ratio2ax],
        "bottom": [0, -ratio2ax - space, 1, ratio2ax],
    }

    ax2 = ax.inset_axes(pos_dic[tick_pos], facecolor="none")
    ax = ax2

    if tick_pos in ["top", "bottom"]:
        ax.set_xlim([start, end])

        xticks = ax.get_xticks()
        xtick_labels = xticks

        last_label_ix = -1
        if start > end:
            last_label_ix = 0

        if scale_adjust == "Mb":
            if end % 1000000 > 0:
                last_label_ix = -2
                if start > end:
                    last_label_ix = 1
            xtick_labels = xtick_labels / 1000000
            xtick_labels = ["{0}".format(tick_fl % i) for i in xtick_labels]
            ax.set_xticks(
                xticks, xtick_labels, fontsize=tick_fontsize, rotation=tick_rotation
            )
            ax.spines["bottom"].set_position(("data", 0))
            # ax.text(end-(abs(end-start)*0.05), -1, 'Mb', fontsize=chrom_fontsize)
        elif scale_adjust == "kb":
            if end % 1000 > 0:
                last_label_ix = -2
                if start > end:
                    last_label_ix = 1
            xtick_labels = xtick_labels / 1000
            xtick_labels = ["{0}".format(tick_fl % i) for i in xtick_labels]
            ax.set_xticks(xticks, xtick_labels)
            # ax.text(end-(abs(end-start)*0.05), -1, 'kb', fontsize=chrom_fontsize)
            ax.spines["bottom"].set_position(("data", 0))
        else:
            pass

        spines = ["bottom", "top", "right", "left"]
        # chrom_x = start + (end-start)/2
        chrom_x = start
        chrom_y = 1
        va = "top"
        ha2 = "right"
        if tick_pos == "bottom":
            del spines[1]
            ax.tick_params(top=True, labeltop=True, bottom=False, labelbottom=False)
            chrom_y = 0
            va = "bottom"

        if tick_pos == "top":
            ha2 = "left"
            del spines[0]
            ax.tick_params(top=False, labeltop=False, bottom=True, labelbottom=True)
        for i in spines:
            ax.spines[i].set_visible(False)
        if tick_rotation == 0:
            ha2 = "center"

        ax.text(chrom_x, chrom_y, raw_region, fontsize=label_fontsize, ha="left", va=va)
        labels = [label.get_text() for label in ax.get_xticklabels()]

        if tick_rotation != 0:
            labels[last_label_ix] = "{0}\n({1})".format(
                labels[last_label_ix], scale_adjust
            )
        else:
            labels[last_label_ix] = labels[last_label_ix] + "({0})".format(scale_adjust)
        ax.set_xticklabels(
            labels, fontsize=tick_fontsize, rotation=tick_rotation, va="center", ha=ha2
        )
        ax.set_xlim([start, end])
        ax.tick_params(which="major", direction="in", pad=-16)
        ax.set_yticks([])
        ax.set_yticklabels("")


def multi_scale_track(
    ax: Optional[Axes] = None,
    regions: Union[Sequence[str], str, None] = None,
    colors: Union[Sequence[str], None] = None,
    alpha: Union[float, None] = 1,
    intervals: Union[int, None] = 1,
    scale_adjust: Union[str, None] = "kb",
    tick_fl: Union[str, None] = "%0.2f",
    tick_fontsize: Union[int, None] = 8,
    tick_rotation: Union[int, None] = 0,
):
    """
    Plot region scale bar, support for multiple or reverse genome regions.

    Parameters
    ----------
    ax: :class:`matplotlib.axes.Axes` object
    regions: `str list` | `str`
        e.g. ``"chr6:1000000-2000000"`` or ``["chr6:1000000-2000000", "chr3:5000000-4000000"]``
        The start can be larger than the end (eg. ``"chr6:2000000-1000000"``),
        which means you want to get the reverse region
    colors: `str list`
        scale bar colors
    alpha: `float`
        scale bar alpha
    intervals: `int`
        rows of the scale bar by region
    scale_adjust: `str`
        options in ['kb', 'Mb']
    tick_fl: `str`
        ticks retains a few decimal places
    tick_fontsize: `int`
        ticks text fontsize
    tick_rotation: `int`
        ticks text rotation

    Example
    -------
    >>> import trackc as tc
    >>> regions = ['7:153000000-151000000', '11:118500000-116500000']
    >>> ten = tc.tenon(figsize=(8,1))
    >>> ten.add(pos='bottom', height=1)
    >>> tc.pl.multi_scale_track(ten.axs(0), regions=regions, scale_adjust='Mb', intervals=2)
    """

    if colors != None:
        track_colors = colors
    else:
        track_colors = trackcl_11

    if isinstance(regions, list):
        line_GenomeRegions = pd.concat(
            [GenomeRegion(i).GenomeRegion2df() for i in regions]
        )
    else:
        line_GenomeRegions = GenomeRegion(regions).GenomeRegion2df()

    line_GenomeRegions = line_GenomeRegions.reset_index()

    if len(track_colors) < line_GenomeRegions.shape[0]:
        repeat_times = (line_GenomeRegions.shape[0] + len(track_colors) - 1) // len(
            track_colors
        )
        track_colors = (track_colors * repeat_times)[: line_GenomeRegions.shape[0]]
    else:
        track_colors = track_colors[: line_GenomeRegions.shape[0]]
    line_GenomeRegions["color"] = track_colors
    line_GenomeRegions["len"] = (
        line_GenomeRegions["fetch_end"] - line_GenomeRegions["fetch_start"]
    )
    line_GenomeRegions["acum"] = line_GenomeRegions["len"].cumsum()
    sum_len = line_GenomeRegions["len"].sum()
    bottom = list(range(intervals))
    # broadcast
    if intervals < line_GenomeRegions.shape[0]:
        repeat_times = (line_GenomeRegions.shape[0] + intervals - 1) // intervals
        bottom = (bottom * repeat_times)[: line_GenomeRegions.shape[0]]
    else:
        bottom = bottom[: line_GenomeRegions.shape[0]]
    line_GenomeRegions.loc[:, "bottom"] = bottom

    line_GenomeRegions["tick_s"] = _scale_ticks(
        line_GenomeRegions["start"], scale=scale_adjust, tick_fl=tick_fl
    )
    line_GenomeRegions["tick_e"] = _scale_ticks(
        line_GenomeRegions["end"], scale=scale_adjust, tick_fl=tick_fl
    )
    if scale_adjust in ["Mb", "kb"]:
        if intervals > 1:
            line_GenomeRegions.loc[line_GenomeRegions.index[-1], "tick_e"] = (
                line_GenomeRegions.loc[line_GenomeRegions.index[-1], "tick_e"]
                + "("
                + scale_adjust
                + ")"
            )
        else:
            line_GenomeRegions["tick_e"] = (
                line_GenomeRegions["tick_e"] + "(" + scale_adjust + ")"
            )

    for i, row in line_GenomeRegions.iterrows():
        arrow_s = row["acum"] - row["len"]
        dx = row["len"]
        if row["isReverse"] == True:
            arrow_s = row["acum"]
            dx = -1 * row["len"]

        ax.arrow(
            arrow_s,
            row["bottom"],
            dx,
            0,
            overhang=1,
            width=0.02,
            head_width=0.3,
            head_length=sum_len / 100,
            length_includes_head=True,
            color=row["color"],
            linewidth=1,
            alpha=alpha,
        )

        tick_ha = "center"
        if tick_rotation != 0:
            tick_ha = "right"
        if intervals > 1:
            ax.text(
                row["acum"] - row["len"] / 2,
                row["bottom"] + 0.1,
                row["chrom"],
                ha="center",
                va="bottom",
                fontsize=tick_fontsize,
                color=row["color"],
            )
            ax.text(
                row["acum"] - row["len"],
                row["bottom"] - 0.1,
                row["tick_s"],
                ha=tick_ha,
                va="top",
                fontsize=tick_fontsize,
                rotation=tick_rotation,
                color=row["color"],
            )
            ax.text(
                row["acum"],
                row["bottom"] - 0.1,
                row["tick_e"],
                ha=tick_ha,
                va="top",
                fontsize=tick_fontsize,
                rotation=tick_rotation,
                color=row["color"],
            )
        else:
            ax.text(
                row["acum"] - row["len"] / 2,
                row["bottom"] - 0.2,
                "{0}:{1}-{2}".format(row["chrom"], row["tick_s"], row["tick_e"]),
                ha="center",
                va="top",
                fontsize=tick_fontsize,
                color=row["color"],
            )

    ax.set_xlim([0, sum_len])
    ax.set_ylim([-0.7, intervals - 0.3])

    for i in ["bottom", "top", "right", "left"]:
        ax.spines[i].set_color("none")
        ax.spines[i].set_linewidth(0)

    ax.tick_params(bottom=False, top=False, left=False, right=False)
    ax.set_xticklabels("")
    ax.set_yticklabels("")
    # ax.spines["bottom"].set_color('black')
    # ax.spines["bottom"].set_linewidth(1)



from typing import Optional, Sequence, Union

import pandas as pd
from matplotlib.axes import Axes

from trackc.tl._getRegionsCmat import GenomeRegion

from .Virtual4C import _get_pets


def _mark_contact_regions_square(
    ax,
    markRegions,
    plot_region_df,
    map_order,
    symmetric,
    linewidth,
    linestyle,
    linecolor,
    only_cis,
):
    for i, reg in enumerate(markRegions):
        reg_pos = plot_region_df[plot_region_df[reg] == 1]
        x0 = reg_pos.index[0] - 0.5
        x1 = reg_pos.index[-1] + 0.5

        """
        ---> ln1
        |ln3 |ln2
        ---->ln4 
        
        (pos1)   _____ (pos4)
                |     |
                |     |
        (pos2)   _____ (pos3)
        """
        (ln1,) = ax.plot(
            (x0, x1),
            (x0, x0),
            linestyle=linestyle,
            color=linecolor,
            linewidth=linewidth,
            solid_capstyle="butt",
        )
        (ln2,) = ax.plot(
            (x1, x1),
            (x0, x1),
            linestyle=linestyle,
            color=linecolor,
            linewidth=linewidth,
            solid_capstyle="butt",
        )
        (ln3,) = ax.plot(
            (x0, x0),
            (x0, x1),
            linestyle=linestyle,
            color=linecolor,
            linewidth=linewidth,
            solid_capstyle="butt",
        )
        (ln4,) = ax.plot(
            (x0, x1),
            (x1, x1),
            linestyle=linestyle,
            color=linecolor,
            linewidth=linewidth,
            solid_capstyle="butt",
        )

        if symmetric == False:
            if map_order == 0:
                ln3.remove()
                ln4.remove()
            else:
                ln1.remove()
                ln2.remove()

        if only_cis == False:
            for region in markRegions[i + 1 :]:
                trans_reg_pos = plot_region_df[plot_region_df[region] == 1]
                y0 = trans_reg_pos.index[0]
                y1 = trans_reg_pos.index[-1]
                (ln5,) = ax.plot(
                    (x0, x1),
                    (y0, y0),
                    linestyle=linestyle,
                    color=linecolor,
                    linewidth=linewidth,
                    solid_capstyle="butt",
                )
                (ln6,) = ax.plot(
                    (x0, x1),
                    (y1, y1),
                    linestyle=linestyle,
                    color=linecolor,
                    linewidth=linewidth,
                    solid_capstyle="butt",
                )

                (ln7,) = ax.plot(
                    (y0, y0),
                    (x0, x1),
                    linestyle=linestyle,
                    color=linecolor,
                    linewidth=linewidth,
                    solid_capstyle="butt",
                )
                (ln8,) = ax.plot(
                    (y1, y1),
                    (x0, x1),
                    linestyle=linestyle,
                    color=linecolor,
                    linewidth=linewidth,
                    solid_capstyle="butt",
                )

                if symmetric == False:
                    if map_order == 1:
                        ln7.remove()
                        ln8.remove()
                    if map_order == 0:
                        ln5.remove()
                        ln6.remove()
            for region in markRegions[:i]:
                trans_reg_pos = plot_region_df[plot_region_df[region] == 1]
                y0 = trans_reg_pos.index[0]
                y1 = trans_reg_pos.index[-1]
                (ln5,) = ax.plot(
                    (y0, y0),
                    (x0, x1),
                    linestyle=linestyle,
                    color=linecolor,
                    linewidth=linewidth,
                    solid_capstyle="butt",
                )
                (ln6,) = ax.plot(
                    (y1, y1),
                    (x0, x1),
                    linestyle=linestyle,
                    color=linecolor,
                    linewidth=linewidth,
                    solid_capstyle="butt",
                )

                (ln7,) = ax.plot(
                    (x0, x1),
                    (y0, y0),
                    linestyle=linestyle,
                    color=linecolor,
                    linewidth=linewidth,
                    solid_capstyle="butt",
                )
                (ln8,) = ax.plot(
                    (x0, x1),
                    (y1, y1),
                    linestyle=linestyle,
                    color=linecolor,
                    linewidth=linewidth,
                    solid_capstyle="butt",
                )

                if symmetric == False:
                    if map_order == 1:
                        ln7.remove()
                        ln8.remove()
                    if map_order == 0:
                        ln5.remove()
                        ln6.remove()


def _mark_contact_regions_triangle(
    ax,
    markRegions,
    plot_region_df,
    map_order,
    symmetric,
    linewidth,
    linestyle,
    linecolor,
    only_cis,
):
    for i, reg in enumerate(markRegions):
        reg_pos = plot_region_df[plot_region_df[reg] == 1]

        x0 = reg_pos.index[0]
        x1 = reg_pos.index[-1]
        (ln1,) = ax.plot(
            (x0, x0 + (x1 - x0) / 2),
            (0, x1 - x0),
            linestyle=linestyle,
            color=linecolor,
            linewidth=linewidth,
            solid_capstyle="butt",
        )
        (ln2,) = ax.plot(
            (x0 + (x1 - x0) / 2, x1),
            (x1 - x0, 0),
            linestyle=linestyle,
            color=linecolor,
            linewidth=linewidth,
            solid_capstyle="butt",
        )

        (ln3,) = ax.plot(
            (x0, x0 + (x1 - x0) / 2),
            (0, (x1 - x0) * -1),
            linestyle=linestyle,
            color=linecolor,
            linewidth=linewidth,
            solid_capstyle="butt",
        )
        (ln4,) = ax.plot(
            (x0 + (x1 - x0) / 2, x1),
            ((x1 - x0) * -1, 0),
            linestyle=linestyle,
            color=linecolor,
            linewidth=linewidth,
            solid_capstyle="butt",
        )
        if symmetric == False:
            if map_order == 0:
                ln3.remove()
                ln4.remove()
            else:
                ln1.remove()
                ln2.remove()

        if only_cis == False:
            for region in markRegions[i + 1 :]:
                trans_reg_pos = plot_region_df[plot_region_df[region] == 1]
                y0 = trans_reg_pos.index[0]
                y1 = trans_reg_pos.index[-1]
                pos_botom = (x1 + (y0 - x1) / 2, y0 - x1)
                pos_top = (x0 + (y1 - x0) / 2, y1 - x0)
                pos_l = (x0 + (y0 - x0) / 2, y0 - x0)
                pos_r = (x1 + (y1 - x1) / 2, y1 - x1)

                (ln5,) = ax.plot(
                    (pos_botom[0], pos_r[0]),
                    (pos_botom[1], pos_r[1]),
                    linestyle=linestyle,
                    color=linecolor,
                    linewidth=linewidth,
                    solid_capstyle="butt",
                )
                (ln6,) = ax.plot(
                    (pos_top[0], pos_l[0]),
                    (pos_top[1], pos_l[1]),
                    linestyle=linestyle,
                    color=linecolor,
                    linewidth=linewidth,
                    solid_capstyle="butt",
                )
                (ln7,) = ax.plot(
                    (pos_botom[0], pos_r[0]),
                    (-pos_botom[1], -pos_r[1]),
                    linestyle=linestyle,
                    color=linecolor,
                    linewidth=linewidth,
                    solid_capstyle="butt",
                )
                (ln8,) = ax.plot(
                    (pos_top[0], pos_l[0]),
                    (-pos_top[1], -pos_l[1]),
                    linestyle=linestyle,
                    color=linecolor,
                    linewidth=linewidth,
                    solid_capstyle="butt",
                )

                (ln5_2,) = ax.plot(
                    (pos_botom[0], pos_l[0]),
                    (pos_botom[1], pos_l[1]),
                    linestyle=linestyle,
                    color=linecolor,
                    linewidth=linewidth,
                    solid_capstyle="butt",
                )
                (ln6_2,) = ax.plot(
                    (pos_top[0], pos_r[0]),
                    (pos_top[1], pos_r[1]),
                    linestyle=linestyle,
                    color=linecolor,
                    linewidth=linewidth,
                    solid_capstyle="butt",
                )
                (ln7_2,) = ax.plot(
                    (pos_botom[0], pos_l[0]),
                    (-pos_botom[1], -pos_l[1]),
                    linestyle=linestyle,
                    color=linecolor,
                    linewidth=linewidth,
                    solid_capstyle="butt",
                )
                (ln8_2,) = ax.plot(
                    (pos_top[0], pos_r[0]),
                    (-pos_top[1], -pos_r[1]),
                    linestyle=linestyle,
                    color=linecolor,
                    linewidth=linewidth,
                    solid_capstyle="butt",
                )

                if symmetric == False:
                    if map_order == 0:
                        ln7.remove()
                        ln8.remove()
                        ln7_2.remove()
                        ln8_2.remove()
                    if map_order == 1:
                        ln5.remove()
                        ln6.remove()
                        ln5_2.remove()
                        ln6_2.remove()


def _mark_contact_regions(
    row_regions,
    markRegions,
    binsize,
    ax,
    map_type,
    map_order,
    symmetric,
    linestyle,
    linecolor,
    linewidth,
    only_cis,
):
    plot_region_df = _get_pets(row_regions, binsize)

    mark_GenomeRegions = pd.concat(
        [GenomeRegion(i).GenomeRegion2df() for i in markRegions]
    )
    mark_GenomeRegions_cp = mark_GenomeRegions.copy()
    for i, row in mark_GenomeRegions_cp.iterrows():
        plot_region_df[i] = 0
        bin_s = int(row["fetch_start"] / binsize)
        if row["fetch_start"] % binsize > 0:
            bin_s = bin_s + 1

        bin_e = int(row["fetch_end"] / binsize)
        if row["fetch_end"] % binsize > 0:
            bin_e = bin_e + 1
        one_region = plot_region_df[plot_region_df["chrom"] == row["chrom"]]
        one_region = one_region[
            (one_region["chrbin"] >= bin_s) & (one_region["chrbin"] <= bin_e)
        ]
        if one_region.shape[0] > 0:
            plot_region_df.loc[one_region.index, i] = 1

    if map_type == "square":
        _mark_contact_regions_square(
            ax,
            markRegions,
            plot_region_df,
            map_order,
            symmetric,
            linewidth,
            linestyle,
            linecolor,
            only_cis,
        )
    elif map_type == "triangle":
        _mark_contact_regions_triangle(
            ax,
            markRegions,
            plot_region_df,
            map_order,
            symmetric,
            linewidth,
            linestyle,
            linecolor,
            only_cis,
        )
    else:
        pass


def mapc_markline(
    ax: Optional[Axes] = None,
    row_regions: Union[pd.DataFrame, None] = None,
    mark_regions: Union[Sequence[str], str, None] = None,
    binsize: Union[int, None] = 10000,
    map_order: int = 0,
    map_type: Union[str, None] = "squ",
    symmetric: bool = False,
    trans_ax: bool = False,
    only_cis: bool = False,
    linestyle: Union[str, None] = "--",
    linecolor: Union[str, None] = "k",
    linewidth: Union[int, None] = 1,
    show_regions_edge: bool = False,
):
    """
    The ``mapc_markline`` function used for mark contact regions on the heatmap,
    It should be used in conjunction with ``mapc``. The parameters "symmetric" and
    "trans_ax" which should be consistent with the parameters used in ``mapc``.

    Parameters
    ----------
    ax: :class:`matplotlib.axes.Axes` object
        utilizes the same `Axes` as used in ``mapc``
    row_regions: `pd.DataFrame`
        The `row_regions` attribute of the object returned by ``tc.tl.extractContactRegions``
        represents the regions extracted from the rows.
    mark_regions: `np.ndarray`
        matrix for plot upper or right of heatmap
    binsize: `int`
        binsize of raw heatmap
    map_order: `int`
        one option of [0, 1]
        The `map_order` parameter in the context of `trackc.mapc` is associated with the `mat` and `mat2` parameters.
        If use `trackc.mapc` to generate a heatmap in the top-right quadrant, then for "map_order" here, you should set it to 0.
        If you use "mapc" to generate a heatmap in the bottom-left quadrant, then for "map_order" here, you should also set it to 0.
    map_type: `str`
        same as the `trackc.mapc` map_type parameter.
        optional is ['square', 'squ', 'triangle', 'tri', 'rectangle', 'rec'], default is `square`
        que is the same as square, tri is the same as triangle, rec is the same as rectangle.
    symmetric: `bool`
        same as the `trackc.mapc` symmetric parameter.
    trans_ax: `bool`
        same as the `trackc.mapc` trans_ax parameter.
    only_cis: `bool`
        If set to True, the trans regions will not be displayed.
    linestyle: `str`
        mark line style
    linecolor: `str`
        mark line color
    linewidth: `int`
        mark line width
    show_regions_edge: `bool`
        Whether to display the spliced region as input from the `row_regions` parameter

    """
    if map_order == 1:
        for child in ax.get_children():
            if isinstance(child, Axes):
                ax = child
                break

    if map_type == "squ":
        map_type = "square"
    if map_type == "tri":
        map_type = "triangle"
    if map_type == "rec":
        map_type = "rectangle"

    if mark_regions != None:
        _mark_contact_regions(
            row_regions,
            mark_regions,
            binsize,
            ax,
            map_type,
            map_order,
            symmetric,
            linestyle,
            linecolor,
            linewidth,
            only_cis,
        )

    if show_regions_edge == True:
        _plot_contact_regions_line(
            ax,
            row_regions,
            map_order,
            map_type,
            trans_ax,
            linewidth,
            linestyle,
            linecolor,
        )


def _plot_contact_regions_line(
    ax,
    rl_regions,
    map_order=0,
    map_type="triangle",
    trans_ax=False,
    linewidth=1,
    linestyle="--",
    linecolor="k",
):
    if trans_ax == True:
        print("trans_ax should be False")

    xA, yA, xB, yB = 0, 0, 0, 0
    xA2, yA2, xB2, yB2 = 0, 0, 0, 0
    rl_regions = rl_regions.query("chrom == chrom")

    cbins = rl_regions["cbins"].to_list()
    full_len = sum(cbins)
    v_sum = 0
    for i, v in enumerate(cbins):
        v_sum = v_sum + v

        if i == len(cbins) - 1:
            break

        # v2 = cbins[i+1]

        if map_type == "triangle":
            if map_order == 0:
                xA, yA, xB, yB = v_sum / 2, v_sum, v_sum, 0
                xA2, yA2, xB2, yB2 = (
                    v_sum,
                    0,
                    v_sum + (full_len - v_sum) / 2,
                    full_len - v_sum,
                )
            else:
                xA, yA, xB, yB = v_sum / 2, -v_sum, v_sum, 0
                xA2, yA2, xB2, yB2 = (
                    v_sum,
                    0,
                    v_sum + (full_len - v_sum) / 2,
                    -(full_len - v_sum),
                )

        if map_type == "square":
            if map_order == 0:
                xA, yA, xB, yB = v_sum - 0.5, -0.5, v_sum - 0.5, v_sum - 0.5
                xA2, yA2, xB2, yB2 = (
                    v_sum - 0.5,
                    v_sum - 0.5,
                    full_len - 0.5,
                    v_sum - 0.5,
                )
            else:
                xA, yA, xB, yB = -0.5, v_sum - 0.5, v_sum - 0.5, v_sum - 0.5
                xA2, yA2, xB2, yB2 = (
                    v_sum - 0.5,
                    v_sum - 0.5,
                    v_sum - 0.5,
                    full_len - 0.5,
                )

        ax.plot(
            [xA, xB],
            [yA, yB],
            linestyle,
            linewidth=linewidth,
            color=linecolor,
            solid_capstyle="butt",
        )
        ax.plot(
            [xA2, xB2],
            [yA2, yB2],
            linestyle,
            linewidth=linewidth,
            color=linecolor,
            solid_capstyle="butt",
        )



import sys
from typing import Optional, Sequence, Union

import matplotlib.pyplot as plt
import pandas as pd
from matplotlib.axes import Axes
from matplotlib.colors import LogNorm  # ,CenteredNorm, SymLogNorm,PowerNorm, Normalize
from matplotlib.colors import Colormap
from mpl_toolkits.axes_grid1.inset_locator import inset_axes

from trackc.pa import fruitpunch, trackcl_9, trackcl_11, zeileis_28
from trackc.tl._getRegionsCmat import GenomeRegion


def bdgmat_track(
    ax: Optional[Axes] = None,
    bed: Union[pd.DataFrame, str, None] = None,
    regions: Union[Sequence[str], str, None] = None,
    style: Union[str, None] = "heatmap",
    logdata: bool = False,
    minrange: Optional[float] = None,
    maxrange: Optional[float] = None,
    color: Union[Sequence[str], None] = "silver",
    cmap: Union[Colormap, str, None] = fruitpunch,
    alpha: Union[float, None] = 1,
    label: Union[str, None] = None,
    label_fontsize: Union[int, None] = 8,
    ticks_on: bool = True,
    tick_fontsize: Optional[int] = 7,
    tick_fl: Optional[str] = "%0.2f",
    line_lg_ncol: Optional[int] = 1,
):
    """\
    Plot bedGraph matrix track, support for multiple or reverse genome regions.
    bedgraph matrix file like bedgraph, columns from 4th are values
    bedgraph matrix should be sorted by chrom, start

    Parameters
    
    ----------
    ax: :class:`matplotlib.axes.Axes` object
    bed: `pd.DataFrame` | `str`
        If ``bed`` if a filepath, the file should have headers, The first three columns are [chrom start end]
        bedGraph matrix format: [chrom start end scorename1 scorename2 ...]
    regions: `str` | `str list`
        The genome regions to plot
        e.g. ``"chr6:1000000-2000000"`` or ``["chr6:1000000-2000000", "chr3:5000000-4000000"]``
        The start can be larger than the end (eg. ``"chr6:2000000-1000000"``), 
            which means you want to get the reverse region
    style: `str`
        heatmap or line, default heatmap
    logdata: `bool` | `bool list`
        whether log the data before plotting
    minrange: `float` | `float list`
        the minimum range of values used to show
    maxrange: `float` | `float list` 
        the maximum range of values used to show
    color: `str` or `list`
        the color of the plot for style: line, if color is a list, the line color will be set by matrix columns order
    cmap: `str` | `matplotlib.colors.Colormap`
        the colormap of the plot for style: heatmap
    alpha: `float`
        alpha of plot color
    label: `str`
        the title of the track, will show on the left
    label_fontsize: `int`
        the label text fontsize
    ticks_on: `bool`
        whether show the ticks
    tick_fontsize: `int`
        ticks text fontsize
    tick_fl: `str`
        ticks retains a few decimal places
    line_lg_ncol: `int`
        the number of columns of legend for style: line
    """

    if isinstance(regions, list):
        line_GenomeRegions = pd.concat(
            [GenomeRegion(i).GenomeRegion2df() for i in regions]
        )
    else:
        line_GenomeRegions = GenomeRegion(regions).GenomeRegion2df()

    # axs = _make_multi_region_ax(ax, line_GenomeRegions)
    line_GenomeRegions = line_GenomeRegions.reset_index()

    if isinstance(bed, str) == True:
        bed = pd.read_table(bed, sep="\t", header=0)

    if style == "line":
        if isinstance(color, list) == False:
            color = [color]
        score_columns = bed.shape[1] - 3
        if len(color) < score_columns:
            # repeat_times = (line_GenomeRegions.shape[0] + len(color) - 1) // len(color)
            # color = (color * repeat_times)[:score_columns]
            print("color sizes less than plots, use defaults")
            if score_columns <= 9:
                color = trackcl_9
            elif score_columns <= 11:
                color = trackcl_11
            elif score_columns <= 28:
                color = zeileis_28
            else:
                pass

    bed["chrom"] = bed["chrom"].astype(str)

    df2plot = pd.DataFrame()
    bed = bed.fillna(0)
    chromos = bed["chrom"].unique()
    for ix, row in line_GenomeRegions.iterrows():
        raw_chr = row["chrom"]
        if row["chrom"] not in chromos:
            if row["chrom"].startswith("chr"):
                row["chrom"] = row["chrom"].lstrip("chr")
            else:
                row["chrom"] = "chr" + row["chrom"]
            if row["chrom"] not in chromos:
                print(f"{raw_chr} not in begmat chroms!")
                # sys.exit(0)
                return

        bed2plot = bed[
            (bed["chrom"] == row["chrom"])
            & (bed["end"] >= row["fetch_start"])
            & (bed["start"] <= row["fetch_end"])
        ]
        if row["isReverse"] == True:
            bed2plot = bed2plot.iloc[::-1]
        if df2plot.shape[0] == 0:
            df2plot = bed2plot
        else:
            df2plot = pd.concat([df2plot, bed2plot], axis=0)

    df2plot = df2plot.iloc[:, 3:]
    if minrange == None:
        minrange = df2plot.min().min()
    if maxrange == None:
        maxrange = df2plot.max().max()

    norm = None
    if logdata:
        norm = LogNorm()
    clim = (minrange, maxrange)

    if style == "heatmap":
        df2plot = df2plot.T
        caxes = ax.pcolormesh(
            df2plot,
            cmap=cmap,
            clim=clim,
            edgecolor="none",
            norm=norm,
            snap=True,
            linewidth=0.001,
        )
        _colorbar(ax, caxes, tick_font_size=tick_fontsize)
        ax.set_xticks([])  # 隐藏刻度
        ax.set_xticklabels([])  # 隐藏刻度标签
        # if
        ax.set_yticks([i + 0.5 for i in range(0, df2plot.shape[0])])  # 隐藏刻度
        if ticks_on:
            ax.set_yticklabels(df2plot.index, fontsize=tick_fontsize)  # 隐藏刻度标签
        else:
            ax.set_yticklabels([])
            ax.set_yticks([])
    elif style == "line":
        # print(df2plot)
        # import sys
        # sys.exit(1)
        df2plot = df2plot.reset_index()
        del df2plot["index"]
        line = ax.plot(df2plot)

        if len(color) < df2plot.shape[1]:
            repeat_times = (df2plot.shape[1] + len(color) - 1) // len(color)
            color = color * repeat_times

        for i, v in enumerate(line):
            v.set_label(df2plot.columns[i])
            v.set(color=color[i])

        ax.set_xlim(0, df2plot.shape[0] - 1)
        print(f"bdg_mat minrange:{minrange} maxrange:{maxrange}")
        ax.set_xticklabels([])
        ax.set_xticks([])
        ax.set_ylim([minrange, maxrange])
        ax.set_yticks([minrange, maxrange])
        ax.set_yticklabels(
            [f"{tick_fl % minrange}", f"{tick_fl % maxrange}"], fontsize=tick_fontsize
        )  # 隐藏刻度标签
        if ticks_on:
            ax.legend(
                loc="best",
                bbox_to_anchor=(1, 0, 0, 1),
                fontsize=tick_fontsize,
                ncol=line_lg_ncol,
            )

    ax.set_ylabel(label, fontsize=label_fontsize)


def _colorbar(axm, im, width="3%", height="100%", tick_font_size=7):
    axins = inset_axes(
        axm,
        width=width,  # width = 5% of parent_bbox width
        height=height,  # height : 50%
        loc="lower left",
        bbox_to_anchor=(1.01, 0, 1, 1),
        bbox_transform=axm.transAxes,
        borderpad=0,
    )
    cbar = plt.colorbar(im, cax=axins, orientation="vertical")
    cbar.set_ticks([im.get_array().min(), im.get_array().max()])
    cbar.ax.tick_params(labelsize=tick_font_size)
    cbar.ax.yaxis.tick_right()
    cbar.ax.spines["top"].set_color("none")
    cbar.ax.spines["right"].set_color("none")
    cbar.ax.spines["bottom"].set_color("none")
    cbar.ax.spines["left"].set_color("none")



import os
from typing import Any, Callable, List, Mapping, Optional, Sequence, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib.axes import Axes
from matplotlib.colors import LogNorm  # ,CenteredNorm, SymLogNorm,PowerNorm, Normalize
from matplotlib.colors import Colormap
from mpl_toolkits.axes_grid1.inset_locator import inset_axes

from trackc.pa import fruitpunch, fruitpunch2

basedir = os.path.abspath(os.path.dirname(__file__))

ColorLike = Union[str, Tuple[float, ...]]


def _check_na_color(
    na_color: Optional[ColorLike], *, img: Optional[np.ndarray] = None
) -> ColorLike:
    if na_color is None:
        if img is not None:
            na_color = (0.0, 0.0, 0.0, 0.0)
        else:
            na_color = "lightgray"
    return na_color


def getData2Map(mat, maxrange=None, minrange=None, trim_range=0.99, inplace=False):
    mat = mat.astype(float)
    # if logdata: mat = np.log2(mat)
    mat[mat == np.inf] = 0.0
    mat[mat == -np.inf] = 0.0
    mat[mat == 0.0] = np.nan

    df = pd.DataFrame(np.ravel(mat))
    df = df[df != np.nan]

    if np.nanmax(mat) > 0:
        if trim_range < 1 and maxrange == None and minrange == None:
            xmaxrange = np.nanpercentile(abs(df), trim_range * 100)
            xminrange = np.nanpercentile(abs(df), 100 - trim_range * 100)
            print("no max min range")
        else:
            if maxrange == None:
                xmaxrange = abs(df).max()[0]
            else:
                xmaxrange = maxrange
            if minrange == None:
                xminrange = abs(df).min()[0]
            else:
                xminrange = minrange

        if inplace == True:
            mat[(mat <= xminrange) & (mat > 0)] = xminrange
            mat[(mat >= xmaxrange) & (mat > 0)] = xmaxrange
            mat[(mat >= -xminrange) & (mat < 0)] = -xminrange
            mat[(mat <= -xmaxrange) & (mat < 0)] = -xmaxrange

        return mat, xmaxrange, xminrange
    else:
        print("Warning: max data <0, no data to plot")


def _plot_pcolormesh(
    mat: np.ndarray,
    ax: Axes,
    norm=None,
    clim=None,
    cmap: Union[Colormap, str, None] = None,
    trans_ax: bool = False,
):
    N = mat.shape[1]
    # Transformation matrix for rotating the heatmap.
    A = np.array([(y, x) for x in range(N, -1, -1) for y in range(N + 1)])
    t = np.array([[1, 0.5], [-1, 0.5]])
    A = np.dot(A, t)

    C = mat
    X = A[:, 1].reshape(N + 1, N + 1)
    Y = A[:, 0].reshape(N + 1, N + 1)

    if trans_ax == False:
        caxes = ax.pcolormesh(
            X,
            Y,
            np.flipud(C),
            cmap=cmap,
            edgecolor="none",
            norm=norm,
            clim=clim,
            snap=True,
            linewidth=0.001,
        )
    else:
        caxes = ax.pcolormesh(
            Y,
            X,
            np.flipud(C),
            cmap=cmap,
            edgecolor="none",
            norm=norm,
            clim=clim,
            snap=True,
            linewidth=0.001,
        )

    return caxes


def _mapC_triview(
    mat: np.ndarray,
    ax: Optional[Axes] = None,
    cmap: Union[Colormap, str, None] = fruitpunch,
    label: Union[str, None] = None,
    label_fontsize: int = 10,
    label_color: Union[str, None] = "k",
    logdata: bool = False,
    trim_range: float = 0.98,
    maxrange: float = None,
    minrange: float = None,
    height: int = 0,
    trans_ax: bool = False,
    map_type: Union[str, None] = "triangle",
    map_order: int = 0,
    symmetric: bool = False,
    k=1,
):
    """
    Draw triangle view of the C data

    Parameters
    mat:

    cmap
    map_type
        contact heatmap type, can be one of ``['square', 'triangle', 'rectangle']``,
        if select ``rectangle`` the ``mapHeight`` parameter should be set.
    mapHeight: int or None
        Parameter for ``map_type: rectangle'``, ``map_type: triangle'`` also can be set.
        it means the heapmap hight bin number

    map_order: int
        0: np.triu()
        1: np.tril()
    k: int
        ``np.tril`` or ``np.triu`` ``k`` parameter

    """
    mat, maxrange, minrange = getData2Map(
        mat, maxrange=maxrange, minrange=minrange, trim_range=trim_range, inplace=False
    )
    print("maxrange:", maxrange, "minrange:", minrange)

    if symmetric == False:
        if map_order == 0:
            mat = np.triu(mat, k=k)
            mat[mat == 0.0] = np.nan
        elif map_order == 1:
            mat = np.tril(mat, k=k)
            mat[mat == 0.0] = np.nan
        else:
            pass

    norm = None
    if logdata:
        norm = LogNorm()

    clim = (minrange, maxrange)
    if map_type == "square":
        im = ax.matshow(mat, cmap=cmap, norm=norm, clim=clim)
    else:
        if map_type == "triangle":
            im = _plot_pcolormesh(
                mat=mat, ax=ax, norm=norm, clim=clim, cmap=cmap, trans_ax=trans_ax
            )

        if map_type == "rectangle":
            im = _plot_pcolormesh(
                mat=mat, ax=ax, norm=norm, clim=clim, cmap=cmap, trans_ax=trans_ax
            )

    _mapC_colorbar(ax, im, trans_ax, map_order, height, map_type)

    return im


def _mapC_label(
    ax,
    label,
    trans_ax,
    map_order,
    map_type,
    height,
    fontsize=10,
    color="k",
    Pair_mat=False,
):
    xmin, xmax = ax.get_xlim()
    ymin, ymax = ax.get_ylim()

    x, y = xmax, ymax
    verticalalignment = "top"
    horizontalalignment = "right"

    if map_type == "square":
        if map_order == 0:
            x, y = xmax, ymax
        else:
            x, y = xmin, ymin
            verticalalignment = "bottom"
            horizontalalignment = "left"
    else:  # triangle, rectangle
        if Pair_mat == False:
            if trans_ax == True:
                ax.set_xlabel(label, fontsize=fontsize, color=color)
            else:
                ax.set_ylabel(label, fontsize=fontsize, color=color)

            return

        if trans_ax == True:
            y = ymax
            if map_order == 0:
                x = xmax
            else:
                x = xmin
                horizontalalignment = "left"
        else:
            x = xmin
            horizontalalignment = "left"
            if map_order == 1:
                y = ymin
                verticalalignment = "bottom"

    bbox_props = dict(boxstyle="round", fc="w", alpha=0.9, pad=0, ec="none")

    ax.text(
        x,
        y,
        label,
        verticalalignment=verticalalignment,
        horizontalalignment=horizontalalignment,
        # transform = ax.transAxes,
        color=color,
        fontsize=fontsize,
        bbox=bbox_props,
    )


def _mapC_colorbar(ax, im, trans_ax, map_order, height, map_type):
    x0, y0, x0_width, y0_height = 0, 1.015, 1, 0.015
    if trans_ax == False:
        x0, y0, x0_width, y0_height = 1.015, 0, 0.015, 1
        if map_order in [0, 1]:
            y0_height = 0.45
            if height > 0:
                x0_width = 0.012
            else:
                x0_width = 0.04
            if map_order == 0:
                y0 = 0.55
            else:
                y0 = 0
        else:
            x0_width = 0.04
            if height > 0:
                x0_width = 0.012
    else:
        if map_order in [0, 1]:
            x0_width = 0.45
            if height > 0:
                y0_height = 0.012
            else:
                y0_height = 0.04
            if map_order == 0:
                x0 = 0.55
            else:
                x0 = 0
        else:
            y0_height = 0.04
            if height > 0:
                y0_height = 0.012

    orientation = "vertical"
    if trans_ax:
        orientation = "horizontal"

    cax = ax.inset_axes([x0, y0, x0_width, y0_height])
    cbar = plt.colorbar(im, ax=ax, cax=cax, orientation=orientation)
    # cbar.set_ticks([im.get_array().min(), im.get_array().max()])
    if trans_ax:
        cbar.ax.xaxis.tick_top()
        cbar.ax.tick_params(axis="x", labelrotation=90)
    else:
        cbar.ax.yaxis.tick_right()
    # cbar.set_label(label, fontsize=8)


def _paraPair(para):
    if isinstance(para, list) == False:
        return [para, para]
    else:
        if len(para) == 1:
            para[1] = para[0]
        else:
            return para
    return para


def mapC(
    ax: Optional[Axes] = None,
    mat: Union[np.ndarray, None] = None,
    mat2: Union[np.ndarray, None] = None,
    cmap: Union[Sequence[Colormap], Sequence[str], Colormap, str, None] = [
        fruitpunch,
        "YlOrRd",
    ],
    label: Union[Sequence[str], str, None] = None,
    label_fontsize: Union[Sequence[int], int] = 10,
    label_color: Union[Sequence[str], str, None] = "k",
    logdata: Union[Sequence[bool], bool] = False,
    maxrange: Union[Sequence[float], float] = None,
    minrange: Union[Sequence[float], float] = None,
    trim_range: Union[Sequence[float], float] = 0.98,
    map_type: Union[str, None] = "triangle",
    height: int = 0,
    trans_ax: bool = False,
    symmetric: bool = False,
    ax_on: bool = True,
    aspect: Union[str, float] = "auto",
):
    """
    Plot contact map, support for multiple or reverse genome regions.
    This function implements the plot method for `np.ndarray`,
    which could get from trackc.tl.extractCisContact or trackc.tl.extractContactRegions
    By default, the trim_range value is fixed so that the 98th percentile (resp. 2th percentile) of each
    interaction matrix is discarded. It therefore allow to remove the extreme values from the matrix,
    mat or mat2 is plotted independently
    If the maxrange parameter is set, data higher that this threshold will be fixed to the maxrange value.

    cmap, label, label_fontsize, label_color, logdata, minrange, maxrange, trim_range, those parameters can be set as a list,
    mat and mat2 will set to the first two values. If those parameters are one single value or the length of list is one,
    then mat and mat2 both set the same value

    Parameters
    ----------
    ax: :class:`matplotlib.axes.Axes` object
    mat: `np.ndarray`
        matrix for plot upper or right of heatmap
    mat2: `np.ndarray`
        matrix for plot bottom or left of heatmap
    cmap: `str` | `matplotlib.colors.Colormap` | `list`
        colormap for continuous annotations, if set as list, mat and mat2 will set to the first two values
    label: `str`
        the title of the track, will show on the left
    label_fontsize: `int`
        the label text fontsize
    label_color: `str`
        the label text color
    logdata: `bool` | `bool list`
        do you want to log the data before plotting the heatmap
    minrange: `float` | `float list`
        the minimum range of values used to define the color palette
    maxrange: `float` | `float list`
        the maximum range of values used to define the color palette
    trim_range: `float` | `float list`
        remove the extreme values by trimming the counts.[0,1]
        define the maxrange and minrange values using the percentile of the interaction matrix
    map_type: `str`
        optional is ['square', 'squ', 'triangle', 'tri', 'rectangle', 'rec'], default is `square`
        que is the same as square, tri is the same as triangle, rec is the same as rectangle.
        `triangle` and `rectangle` default is flip the image 45 degrees to the left.
        for `rectangle` type, the corresponding length of ``height`` will be truncated
        from both ends of the input matrix.
    height: `int`
        if map_type is one of ['triangle', 'rectangle'], `height` indicates the longest interaction bin interval you want to show
    trans_ax: `bool`
        whether flip the image 45 degrees to the right
    symmetric: `bool`
        whether to display a symmetrical heatmap when only one of mat and mat2 is set
    ax_on: `bool`
        whether show the spines
    aspect: `str` | `float`
        optional is 'auto' or 1
        length-width ratio of heatmap

    Example
    -------
    >>> import trackc as tc
    >>> import cooler
    >>> BxPC3 = cooler.Cooler('./BxPC3.chr18.mcool::/resolutions/25000')
    >>> neo_domain_regions = ['18:47950000-48280000', '18:75280000-74850000']
    >>> tumor_zoom = tc.tl.extractContactRegions(clr=BxPC3, row_regions=neo_domain_regions)
    >>> ten = tc.tenon(figsize=(6,1))
    >>> ten.add(pos='bottom', height=0.7, hspace=0.05)
    >>> tc.pl.mapC(ax=ten.axs(0), mat=tumor_zoom.cmat, map_type='tri',
            maxrange=200, minrange=10, label='tumor res=25k', ax_on=False, height=40)
    >>> tc.savefig('trackc_mapc.pdf')
    """

    if map_type == "squ":
        map_type = "square"
    if map_type == "tri":
        map_type = "triangle"
    if map_type == "rec":
        map_type = "rectangle"

    cmap = _paraPair(cmap)
    label = _paraPair(label)
    label_fontsize = _paraPair(label_fontsize)
    label_color = _paraPair(label_color)
    logdata = _paraPair(logdata)
    maxrange = _paraPair(maxrange)
    minrange = _paraPair(minrange)
    trim_range = _paraPair(trim_range)

    k = 1
    k2 = -1

    if isinstance(mat, np.ndarray) == False:
        k2 = 0
    if isinstance(mat2, np.ndarray) == False:
        k = 0

    # ax2 = ax
    Pair_mat = False
    ax2 = ax.inset_axes([0, 0, 1, 1], facecolor="none")
    ax2.set_zorder(1)
    ax2.set_xticklabels([])
    ax2.set_xticks([])
    ax2.set_yticklabels([])
    ax2.set_yticks([])

    if isinstance(mat, np.ndarray) and isinstance(mat2, np.ndarray):
        # ax2 = ax.inset_axes([0, 0, 1, 1], facecolor='none')
        # ax2.set_zorder(-1)
        symmetric = False
        Pair_mat = True

    if isinstance(mat, np.ndarray):
        im = _mapC_triview(
            mat=mat,
            ax=ax,
            cmap=cmap[0],
            label=label[0],
            logdata=logdata[0],
            maxrange=maxrange[0],
            minrange=minrange[0],
            trim_range=trim_range[0],
            height=height,
            trans_ax=trans_ax,
            map_type=map_type,
            map_order=0,
            symmetric=symmetric,
            k=k,
        )

    if isinstance(mat2, np.ndarray):
        # ax2 = ax.inset_axes([0, 0, 1, 1], facecolor='none')
        # ax2.set_zorder(-1)

        im2 = _mapC_triview(
            mat=mat2,
            ax=ax2,
            cmap=cmap[1],
            label=label[1],
            logdata=logdata[1],
            maxrange=maxrange[1],
            minrange=minrange[1],
            trim_range=trim_range[1],
            height=height,
            trans_ax=trans_ax,
            map_type=map_type,
            map_order=1,
            symmetric=symmetric,
            k=k2,
        )

    xylim_conf = pd.read_table(os.path.join(basedir, "mapc_xylim.txt"), header=0)

    if isinstance(mat, np.ndarray) and isinstance(mat2, np.ndarray):
        set_xylim(
            ax, xylim_conf, map_type, 0, symmetric, trans_ax, Pair_mat, mat, height
        )
        set_xylim(
            ax2, xylim_conf, map_type, 1, symmetric, trans_ax, Pair_mat, mat2, height
        )

        ax.set_aspect(aspect)
        ax2.set_aspect(aspect)

    elif isinstance(mat, np.ndarray):
        set_xylim(
            ax, xylim_conf, map_type, 0, symmetric, trans_ax, Pair_mat, mat, height
        )
        ax.set_aspect(aspect)
    elif isinstance(mat2, np.ndarray):
        set_xylim(
            ax, xylim_conf, map_type, 1, symmetric, trans_ax, Pair_mat, mat2, height
        )
        set_xylim(
            ax2, xylim_conf, map_type, 1, symmetric, trans_ax, Pair_mat, mat2, height
        )
        ax.set_aspect(aspect)
        ax2.set_aspect(aspect)
    else:
        pass

    if ax_on == False:
        ax.axis("off")
        ax2.axis("off")

    if isinstance(mat, np.ndarray):
        _mapC_label(
            ax,
            label[0],
            trans_ax,
            0,
            map_type,
            height,
            fontsize=label_fontsize[0],
            color=label_color[0],
            Pair_mat=Pair_mat,
        )
    if isinstance(mat2, np.ndarray):
        _mapC_label(
            ax2,
            label[1],
            trans_ax,
            1,
            map_type,
            height,
            fontsize=label_fontsize[1],
            color=label_color[1],
            Pair_mat=Pair_mat,
        )


def set_xylim(
    axs, conf, map_type, map_order, symmetric, trans_ax, pair_mat, matx, height
):
    conf = conf[
        (conf["map_type"] == map_type)
        & (conf["map_order"] == map_order)
        & (conf["symmetric"] == symmetric)
        & (conf["trans_ax"] == trans_ax)
        & (conf["pair_mat"] == pair_mat)
    ]
    if height == 0:
        height = matx.shape[0]

    conf_values = {
        "mat_w": matx.shape[1],
        "mat_h": matx.shape[0],
        "height": height,
        "height_m": -height,
        "mat_w_height": matx.shape[1] - height,
        "0": 0,
        "-0.5": -0.5,
        "mat_w_0.5m": matx.shape[1] - 0.5,
        "mat_h_0.5m": matx.shape[0] - 0.5,
    }

    if conf.shape[0] == 0:
        print("type not set xylim")
    else:
        ix = conf.index[0]
        xmin = conf_values[conf.loc[ix, "xlim_left"]]
        xmax = conf_values[conf.loc[ix, "xlim_right"]]
        ymin = conf_values[conf.loc[ix, "ylim_bottom"]]
        ymax = conf_values[conf.loc[ix, "ylim_top"]]
        axs.set_xlim([xmin, xmax])
        axs.set_ylim([ymin, ymax])

    axs.set_xticklabels([])
    axs.set_xticks([])
    axs.set_yticklabels([])
    axs.set_yticks([])


def _plot_heatmap_triangle_xticks(
    ax, regin1_binN, regin2_binN, chrom1, start1, end1, chrom2, start2, end2, showXticks
):
    # ax.set_xticks([])
    ax.set_xticks([0, regin1_binN, regin1_binN + regin2_binN])
    ax.set_xticklabels([start1, str(end1) + " " + str(start2), end2])

    ax.set_yticklabels([])
    ax.set_yticks([])

    ax.xaxis.tick_top()
    ax.spines["top"].set_color("k")
    ax.spines["right"].set_color("none")
    ax.spines["bottom"].set_color("none")
    ax.spines["left"].set_color("none")


def _colorbar_triangle(axm, im, ymax, cbar_size=("2%", "10%")):
    height, width = cbar_size
    # if ymax!=None:
    #    height="6%"
    axins1 = inset_axes(
        axm,
        width=width,
        height=height,
        loc=4,
        bbox_to_anchor=(-0.1, 0.7, 1.2, 2.9),
        bbox_transform=axm.transAxes,
    )
    cbar = plt.colorbar(im, cax=axins1, orientation="horizontal")
    cbar.set_ticks([im.get_array().min(), im.get_array().max()])
    cbar.ax.xaxis.tick_top()
    cbar.ax.spines["top"].set_color("none")
    cbar.ax.spines["right"].set_color("none")
    cbar.ax.spines["bottom"].set_color("none")
    cbar.ax.spines["left"].set_color("none")



from trackc.pa.cmaps import colorC, fruitpunch, fruitpunch2, fruitpunch3, washu
from trackc.pa.palettes import (
    default_20,
    default_102,
    trackcl_9,
    trackcl_11,
    vega_10_trackc,
    zeileis_28,
)



from typing import Optional, Sequence, Union

import numpy as np
import pandas as pd
from matplotlib.axes import Axes

from trackc.pl.bigwig import _make_multi_region_ax
from trackc.tl._getRegionsCmat import GenomeRegion


def _two_degree_bc(x_l=10, x_r=90, y_lr=0, y2=10, dots_num=100):
    """
    bezier curve for loop links
    """

    xt = []
    yt = []
    x_mid = (x_l + x_r) / 2
    x_dots12 = np.linspace(x_l, x_mid, dots_num)
    y_dots12 = np.linspace(y_lr, y2, dots_num)
    x_dots23 = np.linspace(x_mid, x_r, dots_num)
    y_dots23 = np.linspace(y2, y_lr, dots_num)
    for i in range(dots_num):
        x = x_dots12[i] + (x_dots23[i] - x_dots12[i]) * i / (dots_num - 1)
        y = y_dots12[i] + (y_dots23[i] - y_dots12[i]) * i / (dots_num - 1)
        xt.append(x)
        yt.append(y)
    return (xt, yt)


def _plot_loop_arc(
    ax, loop_df, color, max_extend, invert_y, start, end, left_anchor, right_anchor
):
    if loop_df.shape[0] == 0:
        return

    top_y = 0

    for i, row in loop_df.iterrows():
        top = row["length"] / max_extend
        if top < 0.5:
            top = 0.5
        elif top > 0.8:
            top = 0.8
        else:
            pass

        xt, yt = _two_degree_bc(
            x_l=row[left_anchor], x_r=row[right_anchor], y_lr=0, y2=top, dots_num=100
        )

        ax.plot(xt, yt, color=color, linewidth=0.5, solid_capstyle="butt")
        if max(yt) > top_y:
            top_y = max(yt)

    ax.set_xlim(start, end)
    if invert_y == True:
        ax.set_ylim(0.5, 0)
    else:
        ax.set_ylim(0, 0.5)


def links_track(
    ax: Optional[Axes] = None,
    data: Union[pd.DataFrame, str, None] = None,
    regions: Union[Sequence[str], str, None] = None,
    links_type: Union[str, None] = "arc",
    color: Union[Sequence[str], None] = "#66AC84",
    maxrange: Union[int, None] = 3000000,
    invert_y: Optional[bool] = False,
    anchor: Union[str, None] = "inside",
    label: Optional[str] = None,
    label_rotation: Union[int, None] = 0,
    label_fontsize: Optional[int] = 12,
    ax_on: bool = False,
):
    """
    Plot loop arc, support for multiple or reverse genome regions.

    Parameters
    ----------
    ax: :class:`matplotlib.axes.Axes` object
    data: `pd.DataFrame` or path of loops file, the file should be no header.
        the file format expected is like this:
        chr1 x1 x2 chr2 y1 y2
        the fields after the y2 will be ignored,
        recommend the result of juicer loop calling
    regions: `str` | `str list`
        The genome regions to show the arc.
        e.g. ``"chr6:1000000-2000000"`` or ``["chr6:1000000-2000000", "chr3:5000000-4000000", "chr5"]``
        The start can be larger than the end (eg. ``"chr6:2000000-1000000"``),
        which means the reverse region
    links_type: `str`
        Optional is ['arc']
    color: `str` or `str list`
        the color of the arc links, if multiple regions, color para can  set as list
    maxrange: `int`
        The maximum distance between two anchor of loop, to filter loops
        if value is None, all loop will bed plotted
    invert_y: `bool`
        whether reverse the y-axis
    anchor: `str`
        Optional is ['inside', 'outside']
        inside: arc link x2 y1
        outside: arc link x2 y1
    label: `str`
        the title of the track, will show on the left
    label_rotation: `int`
        the label text rotation
    label_fontsize: `int`
        the label text fontsize
    ax_on: `bool`
        whether show the spines

    Example
    -------
    >>> import trackc as tc
    >>> regions = ['chr7:153000000-151000000', 'chr11:118500000-116500000']

    >>> ten = tc.tenon(figsize=(8,1))
    >>> ten.add(pos='bottom', height=1)
    >>> ten.add(pos='bottom', height=1, hspace=0.1)
    >>> ten.add(pos='bottom', height=0.4, hspace=0.1)

    >>> loops='/path/ENCFF041XLP.bedpe'
    >>> tc.pl.links_track(ax=ten.axs(0), data=loops, label='GM12878', regions=regions,
                color=['#66AC84', 'tab:purple'], maxrange=3000000, anchor='inside')
    >>> tc.pl.links_track(ax=ten.axs(1), data=loops, label='GM12878', regions=regions,
                color='tab:purple', invert_y=True, anchor='outside', ax_on=True)
    >>> tc.pl.multi_scale_track(ten.axs(2), regions=regions, scale_adjust='Mb', intervals=1, tick_rotation=0, tick_fontsize=10, colors=['#66AC84', 'tab:purple'])
    >>> tc.savefig('trackc_links_track.pdf')
    """

    if isinstance(regions, list):
        line_GenomeRegions = pd.concat(
            [GenomeRegion(i).GenomeRegion2df() for i in regions]
        )
    else:
        line_GenomeRegions = GenomeRegion(regions).GenomeRegion2df()

    axs = _make_multi_region_ax(ax, line_GenomeRegions)
    line_GenomeRegions = line_GenomeRegions.reset_index()

    if isinstance(color, list) == False:
        color = [color]
    if len(color) < line_GenomeRegions.shape[0]:
        repeat_times = (line_GenomeRegions.shape[0] + len(color) - 1) // len(color)
        color = (color * repeat_times)[: line_GenomeRegions.shape[0]]

    if isinstance(data, str) == True:
        data = pd.read_table(data, comment="#", header=None)

    data = data.iloc[:, [0, 1, 2, 3, 4, 5]]
    data.columns = ["chr1", "x1", "x2", "chr2", "y1", "y2"]

    if anchor == "mid":
        left_anchor = "left_mid"
        right_anchor = "right_mid"

    if anchor == "inside":
        left_anchor = "x2"
        right_anchor = "y1"
    if anchor == "outside":
        left_anchor = "x1"
        right_anchor = "y2"

    max_extend = 0
    links_df_list = []

    data["chr1"] = data["chr1"].astype(str)
    data["chr2"] = data["chr2"].astype(str)

    chromos = data["chr1"].unique()
    chrchrom = line_GenomeRegions.loc[0, "chrom"]
    if chrchrom not in chromos:
        if chrchrom.startswith("chr"):
            chrchrom = chrchrom.lstrip("chr")
            if chrchrom in chromos:
                line_GenomeRegions["chrom"] = line_GenomeRegions["chrom"].str.lstrip(
                    "chr"
                )
        else:
            chrchrom = "chr" + chrchrom
            if chrchrom in chromos:
                line_GenomeRegions["chrom"] = "chr" + line_GenomeRegions["chrom"]

    for ix, row in line_GenomeRegions.iterrows():
        loop_bed_plot = data[data["chr1"] == row["chrom"]]

        loop_bed_plot["length"] = loop_bed_plot["y2"] - loop_bed_plot["x1"]

        loop_bed_plot = loop_bed_plot[
            (
                (loop_bed_plot["x1"] >= row["fetch_start"])
                & (loop_bed_plot["x1"] <= row["fetch_end"])
            )
            | (
                (loop_bed_plot["x2"] >= row["fetch_start"])
                & (loop_bed_plot["x2"] <= row["fetch_end"])
                | (loop_bed_plot["y1"] >= row["fetch_start"])
                & (loop_bed_plot["y1"] <= row["fetch_end"])
            )
            | (
                (loop_bed_plot["y2"] >= row["fetch_start"])
                & (loop_bed_plot["y2"] <= row["fetch_end"])
            )
        ]

        if isinstance(maxrange, int):
            loop_bed_plot = loop_bed_plot[
                (loop_bed_plot["y2"] - loop_bed_plot["x1"]) < maxrange
            ]
        # print('looop num:', loop_bed_plot.shape)
        if loop_bed_plot.shape[0] == 0:
            # for i in ['top', 'right', "left", "bottom"]:
            #    ax.spines[i].set_color('none')
            #    ax.spines[i].set_linewidth(0)
            next

        loop_bed_plot["left_mid"] = (loop_bed_plot["x1"] + loop_bed_plot["x2"]) / 2
        loop_bed_plot["right_mid"] = (loop_bed_plot["y1"] + loop_bed_plot["y2"]) / 2

        links_df_list.append(loop_bed_plot)
        if loop_bed_plot.shape[0] > 0:
            if max_extend < max(loop_bed_plot["length"]):
                max_extend = max(loop_bed_plot["length"])

    for ix, row in line_GenomeRegions.iterrows():
        if links_type == "arc":
            _plot_loop_arc(
                axs[ix],
                links_df_list[ix],
                color[ix],
                max_extend,
                invert_y,
                row["start"],
                row["end"],
                left_anchor,
                right_anchor,
            )
        # if links_type == 'loop':
        # plot_triangle()

    ax.set_ylabel(
        label,
        fontsize=label_fontsize,
        rotation=label_rotation,
        horizontalalignment="right",
        verticalalignment="center",
    )

    if ax_on == False:
        spines = ["top", "bottom", "left", "right"]
        if invert_y == True:
            del spines[0]
        else:
            del spines[1]
        for i in spines:
            ax.spines[i].set_visible(False)

    ax.set_xticks([])
    ax.set_xticklabels("")
    ax.set_yticks([])
    ax.set_yticklabels("")
